#include "xAODdataSource/TxAODdataSource.h"
#include "xAODdataSource/TEventCursor.h"
#include "ROOT/TSeq.hxx"
#include <regex>

namespace xAOD {
namespace DataSource {
using TDataSource = ROOT::Experimental::TDF::TDataSource;
using TDataFrame = ROOT::Experimental::TDataFrame;

constexpr auto DEFAULT_TREE_NAME = "CollectionTree";


TxAODdataSource::TxAODdataSource(std::string_view fileNamePattern, std::string_view treeName) :
    m_fileNamePattern(fileNamePattern.data()), m_treeName(treeName.data()) {
    InitMetaData();
}

TxAODdataSource::~TxAODdataSource() {
    m_entryRanges.clear();
    m_columnClassMap.clear();
    m_columnNames.clear();
    m_eventCursors.clear();
}

void TxAODdataSource::SetNSlots(unsigned int nSlots) {
    if (nSlots == 0) { throw std::invalid_argument("Requested number of slots is zero."); }

    m_NSlots = nSlots;
    m_eventCursors.resize(m_NSlots);
    for (auto const slot : ROOT::TSeqU(m_NSlots)){
        auto eventCursor = std::make_unique<TEventCursor>();
        m_eventCursors[slot] = std::move(eventCursor);

        //TEvent expects raw pointer to TChain and it owns the chain.
        std::unique_ptr<TChain> chain = makeDataChain(m_fileNamePattern, m_treeName);
        //Todo: Access mode for TEvent initialization needs to be configurable.
        auto event = std::make_unique<TEvent>(chain.release(), TEvent::EAuxMode::kClassAccess);
        m_eventCursors[slot]->initialise(std::move(event));
    }
    //std::cout << "[SetNSlots] Initialized " << m_NSlots << " slots." << std::endl;
}

void TxAODdataSource::Initialise() {

    const auto sizePerSlot = m_size / m_NSlots;
    auto start = 0UL;
    auto end = 0UL;
    for (auto i : ROOT::TSeqUL(m_NSlots)) {
        start = end;
        end += sizePerSlot;
        m_entryRanges.emplace_back(start, end);
        (void) i;

        //std::cout << "[Initialize] Range" << i << ": [" << start << "-" << end << "]" << std::endl;
    }
    // Todo: redistribute remainder to all slots
    const auto remainder = m_size % m_NSlots;
    m_entryRanges.back().second += remainder;
    std::cout << "xAODdataSource::Initialize Distributed [" << m_size << "] events to [" << m_NSlots << "] slots\n";
    if (remainder > 0) { std::cout << "xAODdataSource::Initialize Appended remainder to last Range\n"; }
}

void TxAODdataSource::Finalise() {
    m_entryRanges.clear();
    for(auto& eventCursor: m_eventCursors){
        eventCursor->finalise();
    }
}

bool TxAODdataSource::HasColumn(std::string_view columnName) const {
    //std::cout << "HasColumn: " << columnName << std::endl;
    return (m_columnClassMap.find(columnName.data()) != m_columnClassMap.end());
}

const std::vector<std::string>& TxAODdataSource::GetColumnNames() const {
    return m_columnNames;
}


std::string TxAODdataSource::GetTypeName(std::string_view columnName) const {
    if (!HasColumn(columnName)) {
        throw std::runtime_error("Requested column does not exist in DataSource.");
    }

    const auto columnClassPair = *(m_columnClassMap.find(columnName.data()));
    return columnClassPair.second;
}

std::vector<std::pair<ULong64_t, ULong64_t>> TxAODdataSource::GetEntryRanges() {
    //TDataFrame uses task stealing for load balancing.
    //The entry ranges therefore are owned/dynamically updated by TDataFrame after this call.
    //DataSource should not assume/store these values during it's lifetime.
    auto ranges(std::move(m_entryRanges));
    return ranges;
}

std::vector<void *>
TxAODdataSource::GetColumnReadersImpl(std::string_view columnName, const std::type_info &typeInfo) {
    if (!HasColumn(columnName)) {
        throw std::runtime_error("Requested column does not exist in DataSource.");
    }

    std::vector<void *> columnReaders(m_NSlots);
    for (auto const slot : ROOT::TSeqU(m_NSlots)) {
        columnReaders[slot] = m_eventCursors[slot]->columnReader(columnName.data(), typeInfo);
    }
    return columnReaders;
}

void TxAODdataSource::SetEntry(unsigned int slot, ULong64_t entry) {
    m_eventCursors[slot]->setEntry(entry);
}

void TxAODdataSource::InitSlot(unsigned int slot, ULong64_t firstEntry) {
    //Todo: Configure workload for given slot
}

void TxAODdataSource::FinaliseSlot(unsigned int slot) {
    //Todo: Reset workload for given slot
}

void TxAODdataSource::InitMetaData() {
    std::unique_ptr<TChain> metaChain = makeDataChain(m_fileNamePattern, m_treeName);
    //Use branch access for TEvent, since only EventFormat information is required for initialisation.
    auto metaEvent = std::make_unique<TEvent>(metaChain.release(), TEvent::kBranchAccess);
    m_size = static_cast<ULong64_t>(metaEvent->getEntries()); //TEvent internally does a range check.
    if (m_size == 0) { throw std::runtime_error("Unable to retrieve event records from event store."); }

    metaEvent->getEntry(0LL);
    auto eventFormat = metaEvent->inputEventFormat();

    EventFormat_v1::const_iterator formatIterator;
    for (formatIterator = eventFormat->begin(); formatIterator != eventFormat->end(); ++formatIterator) {
        const auto& formatInfo = (*formatIterator).second;
        const std::string className{formatInfo.className()};
        const std::string columnName{formatInfo.branchName()};

        m_columnClassMap.emplace(columnName, className);
        m_columnNames.emplace_back(columnName);
    }
}

std::unique_ptr<TChain>
makeDataChain(std::string_view fileNamePattern, std::string_view treeName) {
    std::unique_ptr<TChain> chain;
    {
        R__LOCKGUARD(gROOTMutex);
        chain = std::make_unique<TChain>(treeName.data());
    }
    chain->ResetBit(kMustCleanup);      //Informs ROOT that client will trigger release of chains resources.
    chain->Add(fileNamePattern.data());
    return chain;
}

ROOT::Experimental::TDataFrame
make_xAODdataFrame(std::string_view fileNamePattern, std::string_view treeName = DEFAULT_TREE_NAME) {
    //This function will be called by Analysis user, hence the name follows
    // camel-case format, unlike other methods inherited from ROOT's interface.

    if(fileNamePattern.empty()) { throw std::invalid_argument("Filename pattern is empty."); }
    if(treeName.empty()) { throw std::invalid_argument("Tree name is empty."); }

    TDataFrame tdf(std::make_unique<TxAODdataSource>(fileNamePattern, treeName));
    return tdf;
}

} // end namespace DataSource
} // end namespace xAOD