#include "xAODEventInfo/EventInfo.h"
#include "xAODdataSource/TxAODdataSource.h"
#include "gtest/gtest.h"
#include "gmock/gmock.h"
#include <typeinfo>
#include <AppUtils/TOptions.h>

//TODO: Split this file into sets of tests grouped by functional behavior.
//TODO: Register a common test executable script with atlas_add_test.

namespace xAOD {
namespace DataSource {
namespace Test {

using namespace ::testing;
using TDataSource = ROOT::Experimental::TDF::TDataSource;
using TDataFrame = ROOT::Experimental::TDataFrame;
constexpr auto fileNamePattern = "${DATA}/TestCases/DAOD.HIGG2D2.EventInfo.Electrons.16Events.root";//Todo: Mock files

TEST(MakeDataFrameTest, ThrowOnEmptyFilePattern){
    ASSERT_THROW(make_xAODdataFrame("", AppUtils::DEFAULT_TREE_NAME), std::invalid_argument);
}

TEST(MakeDataFrameTest, DISABLED_ThrowOnInvalidFileName) {
    ASSERT_THROW(make_xAODdataFrame("NoSuchFile", AppUtils::DEFAULT_TREE_NAME), std::runtime_error);
}

TEST(MakeDataFrameTest, ThrowOnEmptyTreeName){
    ASSERT_THROW(make_xAODdataFrame(fileNamePattern, ""), std::invalid_argument);
}

TEST(MakeDataFrameTest, MakeDataFrameWithValidFileNamePatternAndTreeName) {
    ASSERT_NO_THROW(make_xAODdataFrame(fileNamePattern, AppUtils::DEFAULT_TREE_NAME));
}

TEST(MakeDataFrameTest, ValidateSizePerInputFile) {
    TDataFrame xdf = make_xAODdataFrame(fileNamePattern, AppUtils::DEFAULT_TREE_NAME);
    EXPECT_EQ(16ULL /*expectedEventCount*/, xdf.Count().GetValue());
}

TEST(SetNSlotsTest, ThrowOnZeroSlots) {
    TxAODdataSource xds(fileNamePattern, AppUtils::DEFAULT_TREE_NAME);
    ASSERT_THROW(xds.SetNSlots(0), std::invalid_argument);
}

TEST(SetNSlotsTest, SetOneSlots) {
    TxAODdataSource xds(fileNamePattern, AppUtils::DEFAULT_TREE_NAME);
    ASSERT_NO_THROW(xds.SetNSlots(1));
}

TEST(SetNSlotsTest, SetMultipleSlots) {
    TxAODdataSource xds(fileNamePattern, AppUtils::DEFAULT_TREE_NAME);
    ASSERT_NO_THROW(xds.SetNSlots(4));
}

TEST(HasColumnTest, ReturnFalseOnMissingColumn) {
    TxAODdataSource xds(fileNamePattern, AppUtils::DEFAULT_TREE_NAME);
    ASSERT_FALSE(xds.HasColumn("NoSuchColumn"));
}

TEST(HasColumnTest, ReturnTrueOnExistingColumn) {
    TxAODdataSource xds(fileNamePattern, AppUtils::DEFAULT_TREE_NAME);
    ASSERT_TRUE(xds.HasColumn("EventInfo"));
}

//Todo: Redefine this test's expected output. Should the number of columns be on demand or as available in data file?
TEST(GetColumnNamesTest, DISABLED_MultipleColumnNames) {
    const auto expectedVector = std::vector<std::string>{"EventInfo",
                                                         "EventInfoAux.",
                                                         "EventInfoAuxDyn.streamTagDets",
                                                         "EventInfoAuxDyn.streamTagRobs"};
    TxAODdataSource xds(fileNamePattern, AppUtils::DEFAULT_TREE_NAME);
    const auto &columnNames = xds.GetColumnNames();
    ASSERT_THAT(columnNames, SizeIs(expectedVector.size()));
    EXPECT_THAT(columnNames, UnorderedElementsAre("EventInfo",
                                                  "EventInfoAux.",
                                                  "EventInfoAuxDyn.streamTagDets",
                                                  "EventInfoAuxDyn.streamTagRobs"));
}

TEST(GetTypeNameTest, ThrowOnMissingColumn) {
    TxAODdataSource xds(fileNamePattern, AppUtils::DEFAULT_TREE_NAME);
    ASSERT_THROW(xds.GetTypeName("NoSuchColumn"), std::runtime_error);
}

TEST(GetTypeNameTest, ReturnTypeNameOnValidColumnName) {
    TxAODdataSource xds(fileNamePattern, AppUtils::DEFAULT_TREE_NAME);
    ASSERT_THAT(xds.GetTypeName("EventInfo"), StrEq("xAOD::EventInfo_v1"));
}

TEST(GetColumnReadersTest, ThrowOnMissingColumn) {
    TxAODdataSource xds(fileNamePattern, AppUtils::DEFAULT_TREE_NAME);
    ASSERT_THROW(xds.GetColumnReaders<int>("NoSuchColumn"), std::runtime_error);
}

// Validates that columnName and type are consistent.
TEST(GetColumnReadersTest, DISABLED_ThrowOnMissMatchedType) {
    TxAODdataSource xds(fileNamePattern, AppUtils::DEFAULT_TREE_NAME);
    constexpr auto columnName = "EventInfo";
    ASSERT_TRUE(xds.HasColumn(columnName));     // Simulate valid columnName but invalid type.
    ASSERT_THROW(xds.GetColumnReaders<int>(columnName), std::runtime_error);
}

int main(int argc, char **argv) {
    ::testing::InitGoogleTest(&argc, argv);
    return RUN_ALL_TESTS();
}

} // namespace Test
} // namespace DataSource
} // namespace xAOD